import json

import mock
import pandas as pd
import pytest
import responses
from pandas.testing import assert_frame_equal
from responses import matchers

from gantry.exceptions import GantryException, QueryError
from gantry.query.core import dataframe
from gantry.query.core.queryframe import QueryInfo

from ..conftest import END_TIME, MULTIPLE_FEATURES_INT_DATA, ORIGIN, START_TIME


def test_series_filters_property_getter(series_obj_factory, api_client_obj, query_info_obj):
    filters = ["filter-1", "filter-2", "filter-3"]
    parent_dataframe = mock.Mock(api_client=api_client_obj, query_info=query_info_obj)
    parent_dataframe.filters = filters

    series_obj = series_obj_factory("str", parent_dataframe=parent_dataframe)
    assert series_obj.filters == filters


@pytest.mark.parametrize("num_points_kwargs", [{}, {"num_points": 1}, {"num_points": 10}])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_mean(mock_query, num_points_kwargs, series_obj):
    mock_query.return_value = 0.5
    assert series_obj.mean(**num_points_kwargs) == 0.5
    mock_query.assert_called_once_with(
        "mean", num_points=num_points_kwargs.get("num_points", 1), group_by=None
    )


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_mean_error(mock_query, error, series_obj):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj.mean()
    mock_query.assert_called_once_with("mean", num_points=1, group_by=None)


@pytest.mark.parametrize("num_points_kwargs", [{}, {"num_points": 1}, {"num_points": 10}])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_std(mock_query, num_points_kwargs, series_obj):
    mock_query.return_value = 0.6
    assert series_obj.std(**num_points_kwargs) == 0.6
    mock_query.assert_called_once_with(
        "stddev", num_points=num_points_kwargs.get("num_points", 1), group_by=None
    )


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_std_error(mock_query, error, series_obj):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj.std()
    mock_query.assert_called_once_with("stddev", num_points=1, group_by=None)


@pytest.mark.parametrize("num_points_kwargs", [{}, {"num_points": 1}, {"num_points": 10}])
@mock.patch("gantry.query.core.dataframe.GantrySeries.quantile")
def test_median(mock_query, num_points_kwargs, series_obj):
    mock_query.return_value = {0.5: 0.7}
    assert series_obj.median(**num_points_kwargs) == 0.7
    mock_query.assert_called_once_with([0.5], num_points=num_points_kwargs.get("num_points", 1))


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries.quantile")
def test_median_error(mock_query, error, series_obj):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj.median()
    mock_query.assert_called_once_with([0.5], num_points=1)


@pytest.mark.parametrize("num_points_kwargs", [{}, {"num_points": 1}, {"num_points": 10}])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_count(mock_query, num_points_kwargs, series_obj):
    mock_query.return_value = 100
    assert series_obj.count(**num_points_kwargs) == 100
    mock_query.assert_called_once_with(
        "total", num_points=num_points_kwargs.get("num_points", 1), group_by=None
    )


@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_len(mock_query, series_obj):
    mock_query.return_value = 100
    assert len(series_obj) == 100
    mock_query.assert_called_once_with("total", num_points=1, group_by=None)


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_count_error(mock_query, error, series_obj):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj.count()
    mock_query.assert_called_once_with("total", num_points=1, group_by=None)


@pytest.mark.parametrize("num_points_kwargs", [{}, {"num_points": 1}, {"num_points": 10}])
@pytest.mark.parametrize("datatype", ["int", "float"])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_min(mock_query, datatype, num_points_kwargs, series_obj_factory):
    series_obj = series_obj_factory(datatype)
    mock_query.return_value = 1
    assert series_obj.min(**num_points_kwargs) == 1
    mock_query.assert_called_once_with(
        "minimum", num_points=num_points_kwargs.get("num_points", 1), group_by=None
    )


@pytest.mark.parametrize("invalid_datatype", ["str", "bool", "category"])
def test_min_invalid_datatype(invalid_datatype, series_obj_factory):
    with pytest.raises(ValueError):
        series_obj = series_obj_factory(invalid_datatype)
        _ = series_obj.min()


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_min_error(mock_query, error, series_obj):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj.min()
    mock_query.assert_called_once_with("minimum", num_points=1, group_by=None)


@pytest.mark.parametrize("num_points_kwargs", [{}, {"num_points": 1}, {"num_points": 10}])
@pytest.mark.parametrize("datatype", ["int", "float"])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_max(mock_query, datatype, num_points_kwargs, series_obj_factory):
    series_obj = series_obj_factory(datatype)
    mock_query.return_value = 1000
    assert series_obj.max(**num_points_kwargs) == 1000
    mock_query.assert_called_once_with(
        "maximum", num_points=num_points_kwargs.get("num_points", 1), group_by=None
    )


@pytest.mark.parametrize("invalid_datatype", ["str", "bool", "category"])
def test_max_invalid_datatype(invalid_datatype, series_obj_factory):
    with pytest.raises(ValueError):
        series_obj = series_obj_factory(invalid_datatype)
        _ = series_obj.max()


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_max_error(mock_query, error, series_obj):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj.max()
    mock_query.assert_called_once_with("maximum", num_points=1, group_by=None)


@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_histogram(mock_query, series_obj):
    mock_query.return_value = {"a": 10}
    assert series_obj.histogram() == {"a": 10}
    mock_query.assert_called_once_with("category_percents")


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_histogram_error(mock_query, error, series_obj):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj.histogram()
    mock_query.assert_called_once_with("category_percents")


@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_unique(mock_query, series_obj_str):
    mock_query.return_value = {"a": 10, "b": -10, "c": None}
    assert series_obj_str.unique() == set(["a", "b", "c"])
    mock_query.assert_called_once_with("category_percents")


@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_unique_b(mock_query, series_obj_bool):
    mock_query.return_value = {"t": 10, "f": -10}
    assert series_obj_bool.unique() == set(["t", "f"])
    mock_query.assert_called_once_with("category_percents")


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_unique_error(mock_query, error, series_obj_bool):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj_bool.unique()
    mock_query.assert_called_once_with("category_percents")


@pytest.mark.parametrize("vals", [[0.5], [1, 2, 3], []])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_quantile(mock_query, vals, series_obj):
    mock_query.return_value = {1: 10}
    assert series_obj.quantile(vals) == {1.0: 10}
    mock_query.assert_called_once_with("quantiles", quantile_vals=vals)


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_quantile_error(mock_query, error, series_obj):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj.quantile()


@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_pdf(mock_query, series_obj):
    mock_query.return_value = [1, 2, 3]
    pd.testing.assert_frame_equal(series_obj.pdf(), pd.DataFrame([1, 2, 3]))
    mock_query.assert_called_once_with("pdf")


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_pdf_error(mock_query, error, series_obj):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj.pdf()
    mock_query.assert_called_once_with("pdf")


@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_cdf(mock_query, series_obj):
    mock_query.return_value = [1, 2, 3]
    pd.testing.assert_frame_equal(series_obj.cdf(), pd.DataFrame([1, 2, 3]))
    mock_query.assert_called_once_with("cdf")


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_cdf_error(mock_query, error, series_obj):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj.cdf()
    mock_query.assert_called_once_with("cdf")


def test_dtype(series_obj):
    assert series_obj._dtype() == "int"


@pytest.mark.parametrize("tags", [{}, {"foo": "bar"}])
@pytest.mark.parametrize("filters", [[], [{"some": "filter"}]])
@pytest.mark.parametrize(
    "group_by_info",
    [
        (None, "str"),
        ("proj_datanode", "intC"),
        ("feed_datanode", "intB"),
        ("column-missing-from-schema", None),
    ],
)
@mock.patch("gantry.query.core.dataframe._raw_query_cached_request")
def test_aggregate_query(
    mock_query_data, series_obj_factory_with_filters_and_tags, filters, tags, group_by_info
):
    group_by = group_by_info[0]
    group_by_type = group_by_info[1]
    content_dict = {
        "start_time": "2008-09-03T20:56:35",
        "end_time": "2008-09-10T23:58:23",
        "version": "1.2.3",
        "num_points": 1,
        "queries": {
            "query": {
                "query_type": "feature",
                "model_node_id": "ABCD1234",
                "stat": "some-attribute",
                "stat_kwargs": {"some-attribute": {"a": 10}},
                "feature": "series_name",
                "filters": filters,
                "tags": tags,
            }
        },
    }
    if group_by is not None:
        content_dict["queries"]["query"]["group_by"] = [
            {"feature_name": group_by, "dtype": group_by_type}
        ]
        mock_query_data.return_value = {
            "response": "ok",
            "data": {
                "query": [
                    {
                        "group": [{"feature_name": group_by, "value": "group_by_unique_val1"}],
                        "some-attribute": 9,
                    },
                    {
                        "group": [{"feature_name": group_by, "value": "group_by_unique_val2"}],
                        "some-attribute": 11,
                    },
                ]
            },
        }
    else:
        mock_query_data.return_value = {"response": "ok", "data": {"query": {"some-attribute": 10}}}
    content = json.dumps(content_dict, default=lambda x: x.isoformat())

    series_obj = series_obj_factory_with_filters_and_tags("int", filters=filters, tags=tags)

    if group_by_type is not None:
        result = series_obj._aggregate_query("some-attribute", a=10, group_by=group_by)
        mock_query_data.assert_called_with(series_obj.api_client, content, resource="aggregate")
        if group_by is None:
            assert result == 10
        else:
            assert result["some-attribute"].sum() == 20
            assert len(result) == 2
    else:
        with pytest.raises(ValueError, match=f"Cannot find column {group_by} in dataframe."):
            result = series_obj._aggregate_query("some-attribute", a=10, group_by=group_by)


@pytest.mark.parametrize("num_points", [2, 5, 10])
@pytest.mark.parametrize("tags", [{}, {"foo": "bar"}])
@pytest.mark.parametrize("filters", [[], [{"some": "filter"}]])
@mock.patch("gantry.query.core.dataframe._raw_query_cached_request")
def test_aggregate_query_num_points(
    mock_query_data, series_obj_factory_with_filters_and_tags, filters, tags, num_points
):
    content = json.dumps(
        {
            "start_time": "2008-09-03T20:56:35",
            "end_time": "2008-09-10T23:58:23",
            "version": "1.2.3",
            "num_points": num_points,
            "queries": {
                "query": {
                    "query_type": "feature",
                    "model_node_id": "ABCD1234",
                    "stat": "some-attribute",
                    "stat_kwargs": {"some-attribute": {"a": 10}},
                    "feature": "series_name",
                    "filters": filters,
                    "tags": tags,
                }
            },
        },
        default=lambda x: x.isoformat(),
    )
    series_obj = series_obj_factory_with_filters_and_tags("int", filters=filters, tags=tags)
    mock_query_data.return_value = {
        "response": "ok",
        "data": {"query": {"points": [{"A": 10, "B": 10}, {"A": 20, "B": 20}]}},
    }

    res = series_obj._aggregate_query("some-attribute", num_points=num_points, a=10)
    assert_frame_equal(res, pd.DataFrame([{"A": 10, "B": 10}, {"A": 20, "B": 20}]))

    mock_query_data.assert_called_with(series_obj.api_client, content, resource="time_series")


@pytest.mark.parametrize("tags", [{}, {"foo": "bar"}])
@pytest.mark.parametrize("filters", [[], [{"some": "filter"}]])
def test_aggregate_query_error(series_obj_factory_with_filters_and_tags, filters, tags):
    series_obj = series_obj_factory_with_filters_and_tags("int", filters=filters, tags=tags)

    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/aggregate/query".format(ORIGIN),
            json={"response": "bad", "data": {"query": {"some-attribute": 10}}},
            status=400,
        )
        with pytest.raises(
            GantryException,
            match=r"Aggregation query \[some-attribute\] failed: .+",
        ):
            _ = series_obj._aggregate_query("some-attribute", a=10)


@pytest.mark.parametrize("invalid_num_points", [0, -1, -100])
def test_aggregate_query_error_num_points(invalid_num_points, series_obj_factory):
    series_obj = series_obj_factory("int")
    with pytest.raises(
        ValueError,
    ):
        _ = series_obj._aggregate_query("some-attribute", num_points=invalid_num_points, a=10)


def test_raw_query_cached_request_default_resource(api_client_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/aggregate/query".format(ORIGIN),
            json={"response": "ok", "data": "100"},
            match=[matchers.json_params_matcher({"foo": "bar"})],
        )
        assert dataframe._raw_query_cached_request(api_client_obj, '{"foo": "bar"}') == {
            "response": "ok",
            "data": "100",
        }


def test_raw_query_cached_request_provided_resource(api_client_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/foobar/query".format(ORIGIN),
            json={"response": "ok", "data": "100"},
            match=[matchers.json_params_matcher({"foo": "bar"})],
        )
        assert dataframe._raw_query_cached_request(
            api_client_obj, '{"foo": "bar"}', resource="foobar"
        ) == {
            "response": "ok",
            "data": "100",
        }


@pytest.mark.parametrize("op", ["__eq__", "__gt__", "__lt__", "__ge__", "__le__"])
@pytest.mark.parametrize("invalid_filter", [[1, 2, 3], {"foo": "bar"}, object()])
def test_comparator_invalid_filter(invalid_filter, op, series_obj):
    with pytest.raises(TypeError):
        getattr(series_obj, op)(invalid_filter)


@pytest.mark.parametrize("op", ["__eq__", "__gt__", "__lt__", "__ge__", "__le__"])
def test_comparator_invalid_filter_special_case_series(op, series_obj):
    with pytest.raises(TypeError, match=r"Series cannot be compared to each other directly*"):
        getattr(series_obj, op)(series_obj)


@pytest.mark.parametrize(
    ["method", "invalid_dtypes"],
    [
        ("__gt__", ["bool", "str"]),
        ("__ge__", ["bool", "str"]),
        ("__lt__", ["bool", "str"]),
        ("__le__", ["bool", "str"]),
        ("isin", ["bool"]),
        ("contains", ["bool"]),
    ],
)
def test_comparator_invalid_type(method, invalid_dtypes, series_obj_factory):
    for invalid_dtype in invalid_dtypes:
        series_obj = series_obj_factory(invalid_dtype)
        with pytest.raises(ValueError):
            getattr(series_obj, method)(mock.Mock())


@pytest.mark.parametrize(
    ["method", "expected_filter"],
    [
        ("__gt__", "lower_bound"),
        ("__ge__", "inclusive_lower_bound"),
        ("__lt__", "upper_bound"),
        ("__le__", "inclusive_upper_bound"),
    ],
)
def test_valid_sortable_comparators(method, expected_filter, series_obj_factory):
    series_obj = series_obj_factory(
        "int", parent_dataframe=mock.Mock(filters=[], tags={"foo": "bar"})
    )
    value = 10
    result = getattr(series_obj, method)(value)
    assert [f.filter_ for f in result.filter_objs] == [
        {expected_filter: value, "feature_name": "series_name"}
    ]
    assert [f.series.parent_dataframe for f in result.filter_objs] == [series_obj.parent_dataframe]
    assert [f.series.parent_dataframe.tags for f in result.filter_objs] == [{"foo": "bar"}]


def test_isin(series_obj_factory):
    series_obj = series_obj_factory(
        "str", parent_dataframe=mock.Mock(filters=[], tags={"foo": "bar"})
    )
    value = ["foo"]
    result = series_obj.isin(value)
    assert [f.filter_ for f in result.filter_objs] == [
        {"category_query": value, "feature_name": "series_name"}
    ]
    assert [f.series.parent_dataframe for f in result.filter_objs] == [series_obj.parent_dataframe]
    assert [f.series.parent_dataframe.tags for f in result.filter_objs] == [{"foo": "bar"}]


def test_contains(series_obj_factory):
    series_obj = series_obj_factory(
        "str", parent_dataframe=mock.Mock(filters=[], tags={"foo": "bar"})
    )
    value = "foo"
    result = series_obj.contains(value)
    assert [f.filter_ for f in result.filter_objs] == [
        {"string_query": value, "feature_name": "series_name"}
    ]
    assert [f.series.parent_dataframe for f in result.filter_objs] == [series_obj.parent_dataframe]
    assert [f.series.parent_dataframe.tags for f in result.filter_objs] == [{"foo": "bar"}]


@pytest.mark.parametrize("invalid_key", [object(), 1.2, 123, "some-value"])
def test_getitem_invalid_key(invalid_key, series_obj):
    with pytest.raises(ValueError):
        series_obj[invalid_key]


@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=MULTIPLE_FEATURES_INT_DATA,
    new_callable=mock.PropertyMock,
)
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
def test_getitem(get_application_node_id_mock, metadata, series_obj_factory, api_client_obj):
    get_application_node_id_mock.return_value = "12345"
    df = dataframe.GantryDataFrame(
        api_client=api_client_obj,
        application="foobar",
        version="1.2.3",
        env="dev",
        start_time=START_TIME,
        end_time=END_TIME,
        filters=[],
        tags={"foo": "bar"},
    )

    series_obj = series_obj_factory("int", df, name="A")
    assert series_obj.parent_dataframe.tags == {"foo": "bar"}

    filters = dataframe.Filters._from_single_filter({"foo": "bar"}, series_obj)

    new_series = series_obj[filters]
    assert new_series.name == "A"
    assert new_series.parent_dataframe.tags == {"foo": "bar"}


@pytest.mark.parametrize(
    ["type_", "expected"],
    [
        ("bool", "boolean_query"),
        ("str", "string_query"),
        ("uuid", "string_query"),
        ("int", "equals"),
        ("float", "equals"),
    ],
)
def test_eq(type_, expected, series_obj_factory):
    series_obj = series_obj_factory(
        type_, parent_dataframe=mock.Mock(filters=[], tags={"foo" "bar"})
    )
    value = 10
    result = series_obj == value
    assert [f.filter_ for f in result.filter_objs] == [
        {expected: value, "feature_name": "series_name"}
    ]
    assert [f.series.parent_dataframe for f in result.filter_objs] == [series_obj.parent_dataframe]


@pytest.mark.parametrize(
    ["type_", "expected", "other", "value"],
    [
        ("bool", "boolean_query", True, False),
        ("str", "string_query", "str", "^(?!(str)$).*$"),
        ("uuid", "string_query", "str", "^(?!(str)$).*$"),
        ("int", "not_number_query", "not_none", "not_none"),
        ("float", "not_number_query", "not_none", "not_none"),
    ],
)
def test_ne(type_, expected, other, value, series_obj_factory):
    series_obj = series_obj_factory(
        type_, parent_dataframe=mock.Mock(filters=[], tags={"foo": "bar"})
    )
    result = series_obj != other
    assert [f.filter_ for f in result.filter_objs] == [
        {expected: value, "feature_name": "series_name"}
    ]
    assert [f.series.parent_dataframe for f in result.filter_objs] == [series_obj.parent_dataframe]


@pytest.mark.parametrize(
    ["method", "dtype", "query_type", "value"],
    [
        ("__invert__", "bool", "boolean_query", False),
        ("notnull", "any", "null_value", False),
        ("isnull", "any", "null_value", True),
    ],
)
def test_unary(method, dtype, query_type, value, series_obj_factory):
    series_obj = series_obj_factory(
        dtype, parent_dataframe=mock.Mock(filters=[], tags={"foo": "bar"})
    )
    result = getattr(series_obj, method)()
    assert [f.filter_ for f in result.filter_objs] == [
        {query_type: value, "feature_name": "series_name"}
    ]
    assert [f.series.parent_dataframe for f in result.filter_objs] == [series_obj.parent_dataframe]


@pytest.mark.parametrize("dtype", ["str", "int", "float", "other"])
def test_invalid__invert__(dtype, series_obj_factory):
    series_obj = series_obj_factory(dtype)
    with pytest.raises(ValueError):
        series_obj.__invert__()


@pytest.mark.parametrize(["method", "other"], [("__eq__", None), ("__ne__", None)])
def test_other_is_none__eq___and__ne__(method, other, series_obj_factory):
    series_obj = series_obj_factory("any")
    with pytest.raises(ValueError):
        getattr(series_obj, method)(other)


@pytest.mark.parametrize("method", ["__eq__", "__ne__"])
def test_cannot_compare_type__eq___and__ne__(method, series_obj_factory):
    series_obj = series_obj_factory("any")
    with pytest.raises(ValueError):
        getattr(series_obj, method)("not_none")


def test_filter_builder(series_obj_factory):
    series_obj = series_obj_factory(
        "int", parent_dataframe=mock.Mock(filters=[{"bar": "baz"}], tags={"foo": "bar"})
    )
    res = series_obj._filter_builder("foo", "bar")
    assert [f.filter_ for f in res.filter_objs] == [{"feature_name": "series_name", "foo": "bar"}]
    assert [f.series.parent_dataframe for f in res.filter_objs] == [series_obj.parent_dataframe]
    assert series_obj.parent_dataframe.filters == [{"bar": "baz"}]
    assert series_obj.parent_dataframe.tags == {"foo": "bar"}


@pytest.mark.parametrize("n", [1, 5, 10, 11])
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._populate_raw_request")
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._prepare_pandas_dataframe")
def test_head(mock_df, mock_data, n, series_obj_factory):
    series_obj = series_obj_factory(
        "int", parent_dataframe=mock.Mock(filters=[], tags={"foo": "bar"})
    )
    total_num_events = 10
    all_events = list(range(total_num_events))
    mock_data.return_value = all_events[:n]
    mock_df.return_value = pd.DataFrame({"foo": mock_data.return_value})
    pd.testing.assert_frame_equal(series_obj.head(n), mock_df.return_value)
    mock_data.assert_called_with(query_filters=[], ascending=True, n=n, tags={"foo": "bar"})


@pytest.mark.parametrize("n", [1, 5, 10, 11])
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._populate_raw_request")
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._prepare_pandas_dataframe")
def test_tail(mock_df, mock_data, n, series_obj_factory):
    series_obj = series_obj_factory(
        "int", parent_dataframe=mock.Mock(filters=[], tags={"foo": "bar"})
    )
    total_num_events = 10
    all_events = list(range(total_num_events))
    mock_data.return_value = all_events[-n:]
    mock_df.return_value = pd.DataFrame({"foo": mock_data.return_value})
    pd.testing.assert_frame_equal(series_obj.tail(n), mock_df.return_value)
    mock_data.assert_called_with(query_filters=[], ascending=False, n=n, tags={"foo": "bar"})


@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._populate_raw_request")
@mock.patch("gantry.query.core.dataframe.GantryQueryFrame._prepare_pandas_dataframe")
def test_fetch(mock_df, mock_data, series_obj_factory):
    series_obj = series_obj_factory(
        "int", parent_dataframe=mock.Mock(filters=[], tags={"foo": "bar"})
    )
    total_num_events = 10
    all_events = list(range(total_num_events))
    mock_data.return_value = all_events
    mock_df.return_value = pd.DataFrame({"foo": mock_data.return_value})
    pd.testing.assert_frame_equal(series_obj.fetch(), mock_df.return_value)
    mock_data.assert_called_with(query_filters=[], ascending=True, n=None, tags={"foo": "bar"})


@mock.patch("gantry.query.core.dataframe.GantrySeries._describe")
def test_describe(mock_series_describe, series_obj_factory):
    series_obj = series_obj_factory(
        "int",
        parent_dataframe=mock.Mock(
            query_info=QueryInfo("123", "foo", "0", "dev", START_TIME, END_TIME)
        ),
    )
    mock_series_describe.return_value = {"foo": "bar"}

    df = pd.DataFrame(
        ["foo", START_TIME, END_TIME, "bar"],
        index=["application", "from", "to", "foo"],
        columns=["series_name"],
    )
    assert_frame_equal(series_obj.describe(), df)


@pytest.mark.parametrize("dtype", ["int", "float"])
@mock.patch("gantry.query.core.dataframe.GantrySeries.__getitem__")
@mock.patch("gantry.query.core.dataframe.GantrySeries.mean")
@mock.patch("gantry.query.core.dataframe.GantrySeries.max")
@mock.patch("gantry.query.core.dataframe.GantrySeries.min")
@mock.patch("gantry.query.core.dataframe.GantrySeries.quantile")
def test_private_describe_number(
    mock_quantile,
    mock_min,
    mock_max,
    mock_mean,
    mock_get_item,
    dtype,
    series_obj_factory,
):
    series_obj = series_obj_factory(dtype, parent_dataframe=mock.Mock(filters=[]))

    mock_quantile.return_value = {0.25: 0.24, 0.5: 0.4, 0.75: 0.7}
    mock_get_item.return_value = series_obj
    mock_min.return_value = 1
    mock_max.return_value = 10
    mock_mean.return_value = 5

    series_obj.count.side_effect = [100, 1000, 10000]

    assert series_obj._describe() == {
        "count": 100,
        "max": 10,
        "mean": 5,
        "min": 1,
        "not null count": 1000,
        "null count": 10000,
        "q25": 0.24,
        "q50": 0.4,
        "q75": 0.7,
        "series type": "prediction",
        "type": dtype,
    }


@mock.patch("gantry.query.core.dataframe.GantrySeries.__getitem__")
def test_private_describe_bool(
    mock_get_item,
    series_obj_factory,
):
    series_obj = series_obj_factory("bool", parent_dataframe=mock.Mock(filters=[]))

    mock_get_item.return_value = series_obj
    series_obj.count.side_effect = [100, 1000, 10000, 50000]

    assert series_obj._describe() == {
        "count": 100,
        "not null count": 1000,
        "null count": 10000,
        "series type": "prediction",
        "type": "bool",
        "True rate": 500,
    }


@mock.patch("gantry.query.core.dataframe.GantrySeries.__getitem__")
def test_private_describe_str(
    mock_get_item,
    series_obj_factory,
):
    series_obj = series_obj_factory("str", parent_dataframe=mock.Mock(filters=[]))

    mock_get_item.return_value = series_obj
    series_obj.count.side_effect = [100, 1000, 10000]

    assert series_obj._describe() == {
        "count": 100,
        "not null count": 1000,
        "null count": 10000,
        "series type": "prediction",
        "type": "str",
    }


@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=MULTIPLE_FEATURES_INT_DATA,
    new_callable=mock.PropertyMock,
)
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
def test_mix_filtering(get_application_node_id_mock, metadata, series_obj_factory, api_client_obj):
    get_application_node_id_mock.return_value = "12345"
    df = dataframe.GantryDataFrame(
        api_client=api_client_obj,
        application="foobar",
        version="1.2.3",
        env="dev",
        start_time=START_TIME,
        end_time=END_TIME,
        filters=[],
        tags={"foo": "bar"},
    )
    s1 = series_obj_factory("int", df, name="A")
    s2 = series_obj_factory("int", df, name="B")

    series = s1[s2 > 20]
    assert series.name == "A"
    assert series.parent_dataframe.filters == [{"feature_name": "B", "lower_bound": 20}]
    assert series.parent_dataframe.tags == {"foo": "bar"}


@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=MULTIPLE_FEATURES_INT_DATA,
    new_callable=mock.PropertyMock,
)
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
def test_compose_filtering(
    get_application_node_id_mock, metadata, series_obj_factory, api_client_obj
):
    get_application_node_id_mock.return_value = "12345"
    df = dataframe.GantryDataFrame(
        api_client=api_client_obj,
        application="foobar",
        version="1.2.3",
        env="dev",
        start_time=START_TIME,
        end_time=END_TIME,
        filters=[],
        tags={"foo": "bar"},
    )

    s1 = series_obj_factory("int", parent_dataframe=df, name="A")
    s2 = s1[s1 > 20]
    s3 = s2[s2 < 10]

    # < 20 and < 10 filters
    assert s3.parent_dataframe.filters == [
        {"feature_name": "A", "lower_bound": 20},
        {"feature_name": "A", "upper_bound": 10},
    ]
    assert s3.parent_dataframe.tags == {"foo": "bar"}


@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_percent_null(mock_query, series_obj_bool):
    mock_query.return_value = 50.0
    assert series_obj_bool.percent_null() == 50.0
    mock_query.assert_called_once_with("percent_null", num_points=1, group_by=None)


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_percent_null_error(mock_query, error, series_obj_bool):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj_bool.percent_null()
    mock_query.assert_called_once_with("percent_null", num_points=1, group_by=None)


@mock.patch("gantry.query.core.dataframe.GantrySeries._filter_builder")
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=MULTIPLE_FEATURES_INT_DATA,
    new_callable=mock.PropertyMock,
)
@pytest.mark.parametrize("dropna", [True, False])
def test_percent_true(
    metadata,
    get_application_node_id_mock,
    mock_query,
    mock_na,
    series_obj_bool,
    dropna,
    api_client_obj,
):
    get_application_node_id_mock.return_value = "12345"
    df = dataframe.GantryDataFrame(
        api_client=api_client_obj,
        application="foobar",
        version="1.2.3",
        env="dev",
        start_time=START_TIME,
        end_time=END_TIME,
        filters=[],
        tags={"foo": "bar"},
    )

    series_obj_bool.parent_dataframe = df

    filters = dataframe.Filters._from_single_filter({"foo": "bar"}, series_obj_bool)
    mock_query.return_value = 50.0
    mock_na.return_value = filters
    assert series_obj_bool.percent_true(dropna=dropna) == 50.0
    mock_query.assert_called_once_with("percent_true", num_points=1, group_by=None)
    if dropna:
        mock_na.assert_called_once_with("null_value", False)


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_percent_true_error(mock_query, error, series_obj_bool):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj_bool.percent_true()
    mock_query.assert_called_once_with("percent_true", num_points=1, group_by=None)


@mock.patch("gantry.query.core.dataframe.GantrySeries._filter_builder")
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=MULTIPLE_FEATURES_INT_DATA,
    new_callable=mock.PropertyMock,
)
@pytest.mark.parametrize("dropna", [True, False])
def test_percent_false(
    metadata,
    get_application_node_id_mock,
    mock_query,
    mock_na,
    series_obj_bool,
    dropna,
    api_client_obj,
):
    get_application_node_id_mock.return_value = "12345"
    df = dataframe.GantryDataFrame(
        api_client=api_client_obj,
        application="foobar",
        version="1.2.3",
        env="dev",
        start_time=START_TIME,
        end_time=END_TIME,
        filters=[],
        tags={"foo": "bar"},
    )

    series_obj_bool.parent_dataframe = df

    filters = dataframe.Filters._from_single_filter({"foo": "bar"}, series_obj_bool)
    mock_query.return_value = 50.0
    mock_na.return_value = filters
    assert series_obj_bool.percent_false(dropna=dropna) == 50.0
    mock_query.assert_called_once_with("percent_false", num_points=1, group_by=None)
    if dropna:
        mock_na.assert_called_once_with("null_value", False)


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_percent_false_error(mock_query, error, series_obj_bool):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj_bool.percent_false()
    mock_query.assert_called_once_with("percent_false", num_points=1, group_by=None)


@mock.patch("gantry.query.core.dataframe.GantrySeries._filter_builder")
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=MULTIPLE_FEATURES_INT_DATA,
    new_callable=mock.PropertyMock,
)
@pytest.mark.parametrize("dropna", [True, False])
def test_percent_true_not_null(
    metadata,
    get_application_node_id_mock,
    mock_query,
    mock_na,
    series_obj_bool,
    dropna,
    api_client_obj,
):
    get_application_node_id_mock.return_value = "12345"
    df = dataframe.GantryDataFrame(
        api_client=api_client_obj,
        application="foobar",
        version="1.2.3",
        env="dev",
        start_time=START_TIME,
        end_time=END_TIME,
        filters=[],
        tags={"foo": "bar"},
    )

    series_obj_bool.parent_dataframe = df

    filters = dataframe.Filters._from_single_filter({"foo": "bar"}, series_obj_bool)
    mock_query.return_value = 50.0
    mock_na.return_value = filters
    assert series_obj_bool.percent_true_not_null(dropna=dropna) == 50.0
    mock_query.assert_called_once_with("percent_true_not_null", num_points=1, group_by=None)
    if dropna:
        mock_na.assert_called_once_with("null_value", False)


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_percent_true_not_null_error(mock_query, error, series_obj_bool):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj_bool.percent_true_not_null()
    mock_query.assert_called_once_with("percent_true_not_null", num_points=1, group_by=None)


@mock.patch("gantry.query.core.dataframe.GantrySeries._filter_builder")
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
@mock.patch("gantry.query.core.dataframe.get_application_node_id")
@mock.patch(
    "gantry.query.core.queryframe.GantryQueryFrame.metadata",
    return_value=MULTIPLE_FEATURES_INT_DATA,
    new_callable=mock.PropertyMock,
)
@pytest.mark.parametrize("dropna", [True, False])
def test_percent_false_not_null(
    metadata,
    get_application_node_id_mock,
    mock_query,
    mock_na,
    series_obj_bool,
    dropna,
    api_client_obj,
):
    get_application_node_id_mock.return_value = "12345"
    df = dataframe.GantryDataFrame(
        api_client=api_client_obj,
        application="foobar",
        version="1.2.3",
        env="dev",
        start_time=START_TIME,
        end_time=END_TIME,
        filters=[],
        tags={"foo": "bar"},
    )

    series_obj_bool.parent_dataframe = df

    filters = dataframe.Filters._from_single_filter({"foo": "bar"}, series_obj_bool)
    mock_query.return_value = 50.0
    mock_na.return_value = filters
    assert series_obj_bool.percent_false_not_null(dropna=dropna) == 50.0
    mock_query.assert_called_once_with("percent_false_not_null", num_points=1, group_by=None)
    if dropna:
        mock_na.assert_called_once_with("null_value", False)


@pytest.mark.parametrize("error", [AttributeError, ValueError, KeyError])
@mock.patch("gantry.query.core.dataframe.GantrySeries._aggregate_query")
def test_percent_false_not_null_error(mock_query, error, series_obj_bool):
    mock_query.side_effect = error
    with pytest.raises(QueryError):
        series_obj_bool.percent_false_not_null()
    mock_query.assert_called_once_with("percent_false_not_null", num_points=1, group_by=None)



import base64
import collections
import collections.abc
import datetime
import hashlib
import json
import logging
import os
import pprint
import re
import sys
from pathlib import Path
from typing import IO, Any, Dict, List, Optional, Set, Tuple

import click
import requests
from dateutil import parser

from gantry.serializers import EventEncoder

logger = logging.getLogger(__name__)


def to_datetime(s: str) -> datetime.datetime:
    """
    Converts a string to a naive UTC datetime object
    """
    try:
        # fromisoformat is present starting in python3.7
        if s.endswith("Z"):
            dt = datetime.datetime.fromisoformat(s.replace("Z", "+00:00"))
        else:
            dt = datetime.datetime.fromisoformat(s)
    except (AttributeError, ValueError, TypeError):
        try:
            dt = datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%f")
        except Exception:
            dt = parser.parse(s)
    if _is_offset_naive(dt):
        return dt

    naive_dt = dt.astimezone(datetime.timezone.utc).replace(tzinfo=None)
    return naive_dt


def to_isoformat_duration(relative_interval: datetime.timedelta) -> str:
    days = relative_interval.days
    seconds = relative_interval.seconds

    minutes, seconds = seconds // 60, seconds % 60
    hours, minutes = minutes // 60, minutes % 60
    weeks, days = days // 7, days % 7

    timedelta_str = "P"
    if weeks > 0:
        timedelta_str += str(weeks) + "W"
    if days > 0:
        timedelta_str += str(days) + "D"
    timedelta_str += "T"
    if hours > 0:
        timedelta_str += str(hours) + "H"
    if minutes > 0:
        timedelta_str += str(minutes) + "M"
    if seconds > 0:
        timedelta_str += str(seconds) + "S"

    if timedelta_str[-1] == "T":
        timedelta_str = timedelta_str[:-1]

    if len(timedelta_str) == 1:
        return "P0D"

    return timedelta_str


def _is_offset_naive(d: datetime.datetime) -> bool:
    # https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
    aware = d.tzinfo is not None and d.tzinfo.utcoffset(d) is not None
    return not aware


def check_event_time_in_future(event_timestamp: datetime.datetime) -> bool:
    """
    Raises an error if the event timestamp is in the future.
    Note, `event_timestamp` is assumed to be in UTC. This
    method will handle both offset-naive and offset-aware event_timestamp
    """
    current_time = (
        datetime.datetime.utcnow()
        if _is_offset_naive(event_timestamp)
        else datetime.datetime.now(datetime.timezone.utc)
    )
    return event_timestamp > current_time


def compute_feedback_id(inputs: Dict[str, Any], feedback_keys: Optional[List[str]] = None) -> str:
    assert isinstance(inputs, collections.abc.Mapping)
    if not feedback_keys:
        # When not specified, default to feedback_keys are all the
        # fields of inputs
        feedback_keys = list(inputs.keys())

    values = []
    for key in sorted(feedback_keys):
        input_value = inputs[key]
        # TODO change to hashable type if input_value isn't
        # hashable
        values.append(input_value)

    return hashlib.md5(
        json.dumps(values, sort_keys=True, cls=EventEncoder).encode("utf-8")
    ).hexdigest()


def clean_name(name: str) -> str:
    """Replace any non-alphanumeric characters, except '-' and '.', with '-'"""
    return re.sub(r"[^a-zA-Z0-9\-.]+", "-", name)


def generate_gantry_name(name: str, max_len: int = 64) -> str:
    prefix = "gantry-"
    suffix = "-" + hashlib.sha256(name.encode("utf-8")).hexdigest()[:8]

    remaining_len = max_len - len(prefix) - len(suffix)
    trunc_name = clean_name(name)[:remaining_len].lower().strip("-")

    return prefix + trunc_name + suffix


def parse_s3_path(path: str) -> Tuple[str, str]:
    """Parses an S3 path of the form s3://bucket/path/to/obj and returns (key, path/to/obj)"""
    prefix_len = len("s3://")
    s3_path = path[prefix_len:]
    bucket, _, key = s3_path.partition("/")

    return (bucket, key)


def format_msg_with_color(msg: str, color: str, logger: logging.Logger) -> str:
    """Given a message and a colorama color, conditionally returns a formatted message if the only
    handlers of the gantry package logger are NullHandlers or StreamHandlers with sys.stderr
    or sys.stdout
    """
    # loop to short-circuit and return original message if a potential file handler is found
    for handler in logging.getLogger("gantry").handlers:
        if isinstance(handler, logging.NullHandler):
            # It is okay if handler is NullHandler
            continue
        elif isinstance(handler, logging.StreamHandler):
            # If handler is a StreamHandler make sure stream is sys.stderr or sys.stdout
            if handler.stream in (
                sys.stderr,
                sys.stdout,
            ):
                continue
        else:
            # In all other cases, we might be logging to a file so just return the original msg
            return msg
    # If we haven't short-circuited and returned original message, return the colored message
    return color + msg


def obj_by_id(els):
    obj = {}
    for el in els:
        obj[el.id] = el

    return obj


def download_file_from_url(presigned_url: str, local_path: Path):
    """
    Args:
        presigned_url (str): presigned url
        local_path (Path): local file path
    """
    os.makedirs(local_path.parent, exist_ok=True)
    with local_path.open("wb") as f:
        _download_file_to_IO(presigned_url, f)
    return local_path


def _download_file_to_IO(presigned_url: str, file: IO):
    """
    Args:
        presigned_url (str): presigned url
        file (IO): file object
    """
    with requests.get(presigned_url, stream=True) as r:
        r.raise_for_status()
        for chunk in r.iter_content(chunk_size=8192):
            file.write(chunk)


def upload_file_to_url(presigned_url: str, local_path: Path):
    """
    Args:
        presigned_url (str): presigned url
        local_path (Path): local file path

    Returns:
        s3 object version id
    """
    with local_path.open("rb") as f:
        return _upload_file_from_IO(presigned_url, f)


def _upload_file_from_IO(presigned_url: str, file: IO):
    """
    Args:
        presigned_url (str): presigned url
        file (IO): file object

    Returns:
        s3 object version id
    """
    res = requests.put(presigned_url, data=file.read())
    res.raise_for_status()
    return res.headers["x-amz-version-id"]


def is_hidden(path):
    return str(path.name).startswith(".")


def list_all_files(dirpath, include_hidden_file=False):
    """
    List all files in a folder

    Args:
        dirpath (Path): target directory
        include_hidden_file (bool): whether include hidden file or not
    Returns:
        List[Path]
    """
    assert dirpath.is_dir()
    file_list = []

    if not include_hidden_file and is_hidden(dirpath):
        return file_list

    for x in dirpath.iterdir():
        if x.is_file():
            if include_hidden_file or not is_hidden(x):
                file_list.append(x)
        elif x.is_dir():
            file_list.extend(list_all_files(x, include_hidden_file))

    return file_list


def get_s3_sha256_checksum(data_bytes):
    """
    Generate local file base64 encoded sha256 checksum

    Args:
        data_bytes : byte data which need to be hashed

    Returns:
        sha256: base64 encoded sha256 value
    """
    return base64.b64encode(hashlib.sha256(data_bytes).digest()).decode("utf-8")


def get_files_checksum(path):
    """_summary_

    Args:
        path (_type_): directory path

    Returns:
        {file name: sha256 checksum}
    """
    # Do not include hidden files to avoid adding .DS_Store or other system files by accident
    file_list = list_all_files(path)
    return dict(
        [(str(f.relative_to(path)), get_s3_sha256_checksum(f.read_bytes())) for f in file_list]
    )


def pprint_records(records: List[Dict[str, str]], include_keys: Optional[Set[str]] = None):
    """
    Args:
        records (List[Dict[str, str]]): _description_
        include_keys (Optional[Set[str]], optional): _description_. Defaults to None.
    """
    pp = pprint.PrettyPrinter(depth=4)
    if not include_keys:
        pp.pprint(records)
    else:
        pp.pprint(
            [dict([(k, v) for k, v in record.items() if k in include_keys]) for record in records]
        )


def check_response_status(resp, error_prefix: str):
    if resp.get("response") != "ok":
        click.secho("FAILED.", fg="red")
        err_msg = resp.get("error", "unknown error")
        raise click.ClickException(f"{error_prefix}: {err_msg}.")
    click.secho("SUCCESS", fg="cyan")


def read_lines_from_url(url):
    response = requests.get(url, stream=True)
    for line in response.iter_lines():
        yield line.decode("utf-8")


def from_isoformat_duration(duration_string: str) -> datetime.timedelta:
    # Parse the duration string using regular expressions
    match = re.match(r"^P((\d+)W)?((\d+)D)?T?((\d+)H)?((\d+)M)?((\d+)S)?$", duration_string)
    if not match:
        raise ValueError(f"Invalid duration string: {duration_string}")

    # Extract the component values from the match object
    weeks = int(match.group(2) or 0)
    days = int(match.group(4) or 0)
    hours = int(match.group(6) or 0)
    minutes = int(match.group(8) or 0)
    seconds = int(match.group(10) or 0)

    # Calculate the total duration in seconds
    total_seconds = (weeks * 7 + days) * 86400 + hours * 3600 + minutes * 60 + seconds

    # Return a timedelta object representing the duration
    return datetime.timedelta(seconds=total_seconds)



import pytest

HOST = "https://test-api"


@pytest.fixture(scope="function")
def test_curators():
    return [
        {
            "id": "01234567-89ab-cdef-0123-456789abcdef",
            "application_name": "test_curator_app",
            "name": "test_curator_name",
            "curated_dataset_name": "test_dataset",
            "start_on": "2023-01-20T22:36:47+00:00",
            "curation_interval": "12:00:00",
            "curation_delay": "00:00:00",
            "curate_past_intervals": False,
            "created_at": "2024-01-20T22:36:47+00:00",
            "selectors": [],
        },
        {
            "id": "6d4e42bc-ff43-4fe4-8625-e368c07b7f08",
            "application_name": "test_selector_app",
            "name": "test_selector_name",
            "curated_dataset_name": "test_dataset_selector",
            "start_on": "2023-01-01T23:12:41+00:00",
            "curation_interval": "20:00:00",
            "curation_delay": "05:00:00",
            "curate_past_intervals": True,
            "created_at": "2020-05-08T01:12:51+00:00",
            "selectors": [
                {
                    "limit": 10,
                    "method": {"field": "inputs.input1", "sample": "ordered", "sort": "ascending"},
                },
                {
                    "limit": 10,
                    "method": {"sample": "uniform"},
                    "filters": [
                        {"field": "inputs.input1", "upper": 70, "lower": 30},
                        {"field": "inputs.input2", "contains": "text"},
                    ],
                },
            ],
        },
    ]


@pytest.fixture(scope="function")
def test_curator(test_curators):
    return test_curators[0]



import dataclasses
import itertools
import logging
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from typing import Dict, List, Optional, Union
from uuid import UUID

from gantry.api_client import APIClient
from gantry.exceptions import GantryException

logger = logging.getLogger(__name__)


class SlackNotificationFrequency(Enum):
    DAILY = "DAILY"
    IMMEDIATELY = "IMMEDIATELY"


class AlertsAggregation(str, Enum):
    """
    Allowed aggregations for Gantry Alerts.

    The full list of aggregations are:

    .. code-block:: python

        TOTAL
        MAX
        MIN
        MEAN
        STD
        PDF
        QUANTILES
        PERCENT_NULL
        PERCENT_TRUE
        PERCENT_FALSE
        PERCENT_TRUE_NOT_NULL
        PERCENT_FALSE_NOT_NULL
        CATEGORY_PERCENTS
        ACCURACY
        MAE
        MSE
        MAPE
        MAX_ERROR
        CONFUSION_MATRIX
        F1
        R2
        PRECISION
        RECALL
        ROC_AUC_SCORE
        PR_AUC_SCORE
        D1
        DINF
        KL
        KS
        CATEGORICAL_D1
        CATEGORICAL_DINF
        CATEGORICAL_KL
        X2
        ROC_CURVE_PLOT
        PR_CURVE_PLOT
        DISTRIBUTION_SKETCH
        CATEGORICAL_SKETCH

    """

    # Standard Aggregation
    TOTAL = "total"
    MAX = "maximum"
    MIN = "minimum"
    MEAN = "mean"
    STD = "stddev"
    PDF = "pdf"
    QUANTILES = "quantiles"
    PERCENT_NULL = "percent_null"
    PERCENT_TRUE = "percent_true"
    PERCENT_FALSE = "percent_false"
    PERCENT_TRUE_NOT_NULL = "percent_true_not_null"
    PERCENT_FALSE_NOT_NULL = "percent_false_not_null"
    CATEGORY_PERCENTS = "category_percents"
    # Metric Aggregations
    ACCURACY = "accuracy_score"
    MAE = "mean_absolute_error"
    MSE = "mean_squared_error"
    MAPE = "mean_absolute_percentage_error"
    MAX_ERROR = "maximum_error"
    CONFUSION_MATRIX = "confusion_matrix"
    F1 = "f1_score"
    R2 = "r2_score"
    PRECISION = "precision_score"
    RECALL = "recall_score"
    ROC_AUC_SCORE = "roc_auc_score"
    PR_AUC_SCORE = "precision_recall_auc_score"
    D1 = "d1"
    DINF = "dinf"
    KL = "kl"
    KS = "ks"
    CATEGORICAL_D1 = "categorical_d1"
    CATEGORICAL_DINF = "categorical_dinf"
    CATEGORICAL_KL = "categorical_kl"
    X2 = "x2"
    # Plots
    ROC_CURVE_PLOT = "roc_curve_plot"
    PR_CURVE_PLOT = "pr_curve_plot"
    # Sketch Aggregations.
    DISTRIBUTION_SKETCH = "distribution_sketch"
    CATEGORICAL_SKETCH = "categorical_sketch"


@dataclass
class AlertsCheck:
    """
    Data shape definition for alert checks.

    column_names correspond to the schema columns in your application.
    """

    aggregation: AlertsAggregation
    column_names: List[str]
    lower_bound: float
    upper_bound: float
    check_type: str = "range_check"


class GantryAlerts:
    ALERT_TYPE = "RangeAlert"  # only one supported
    CHANNEL_TYPE = "SLACK_WEBHOOK"  # only one supported

    def __init__(self, api_client: APIClient):
        self._api_client = api_client

    @classmethod
    def from_api_key(cls, api_url, api_key):
        return cls(api_client=APIClient(origin=api_url, api_key=api_key))

    def _get_application(self, application_name: str) -> Dict:
        """
        Get all application details for a particular application in Gantry.

        This is useful in case you need the application or column (data node) IDs
        to perform some other interaction with the API for your application.

        Args:
            application_name (str): The name of your existing application in Gantry.

        Returns:
            Dict: All application details in Gantry.
        """
        response = self._api_client.request(
            "GET", f"/api/v1/models/{application_name}/schemas", raise_for_status=True
        )
        return response["data"]

    def _map_column_names_to_ids(
        self, application_name: str, column_names: List[str]
    ) -> Dict[str, str]:
        """
        Returns a map of column_name -> data_node_id.

        Args:
            application_name (str): Application in Gantry.
            column_names (List[str]): Data node names that map to the schema of the application.

        Returns:
            Dict[str, str]: Map from column_name to data_node_id
        """
        application_data = self._get_application(application_name)

        feedback_data_nodes = application_data.get("feedback_datanodes", [])
        prediction_data_nodes = application_data.get("prediction_datanodes", [])
        projection_data_nodes = application_data.get("projection_datanodes", [])

        name_to_id = {}
        for column_name in column_names:
            found = False
            for data_node in itertools.chain(
                feedback_data_nodes, prediction_data_nodes, projection_data_nodes
            ):
                if data_node["name"] == column_name:
                    name_to_id[column_name] = data_node["id"]
                    found = True
                    break
            if not found:
                raise ValueError(f"{column_name} not found in application.")

        return name_to_id

    def get_alerts(
        self,
        application_name: str,
        triggered: bool = False,
        triggered_from: Optional[str] = None,
        triggered_to: Optional[str] = None,
    ) -> List[Dict]:
        """
        Get alerts (and their properties) that exist for your application in Gantry. This
        method also allows you to get alerts that have been triggered within a time window.

        NOTE: The data shape of the returned objects differs slightly if you specify the
        `triggered` filter.

        Args:
            application_name (str): The name of your existing application in Gantry.
            triggered (bool): Whether or not to get alerts that have been triggered by
                their check definitions.
            triggered_from (Optional[str]): If provided, will get triggered alerts from
                this time. ISO formatted string. Defaults to the beginning of the UNIX epoch.
            triggered_to (Optional[str]): If provided, will get triggered alerts to this
                time. ISO formatted string. Defaults to now() + 1 day to capture all recent
                alerts.
        Returns:
            List[Dict]: All of the alerts and their properties that exist under the
                application_name (optionally, that are `triggered`)
        """
        if not triggered:
            if triggered_from or triggered_to:
                raise ValueError(
                    "Triggered time window is only valid when requesting triggered alerts."
                )
            query_params = {"model_name": application_name}
            response = self._api_client.request(
                "GET", "/api/v1/alerts", params=query_params, raise_for_status=True
            )
            return response["data"]["alerts"]
        else:
            triggered_from = triggered_from or datetime.utcfromtimestamp(0).isoformat()
            # Have noticed that for just triggered alerts, there can be a lag, so by default
            # we can pad the window.
            triggered_to = triggered_to or (datetime.now() + timedelta(days=1)).isoformat()
            json_data = {
                "start_time": triggered_from,
                "end_time": triggered_to,
                "statuses": ["OPEN"],  # options are 'OPEN' or 'CLOSED'
            }
            response = self._api_client.request(
                "POST", "/api/v1/alerts/triggered/all", json=json_data, raise_for_status=True
            )
            return response["data"]["triggered_alerts"]

    def _create_or_update_alert(
        self,
        application_name: str,
        alert_name: str,
        alert_checks: List[AlertsCheck],
        evaluation_window: str,
        evaluation_delay: str,
        tags: Optional[Dict[str, str]] = None,
        alert_id: Optional[Union[str, UUID]] = None,
    ) -> str:
        """
        Create or update (upsert) an alert in the Gantry application.

        Example check:

        .. code-block:: python

            from gantry.alerts import AlertsCheck

            check = AlertsCheck(
                aggregation=AlertsAggregation.MAX,
                data_node_names=[
                    "inputs.field",
                ],
                lower_bound=10,
                upper_bound=20,
            )

        Args:
            application_name (str): The name of your existing application in Gantry.
            alert_name (str): Whatever you would like to call your (existing) alert!
            alert_checks (List[Dict]): A list of metric checks that define this alert.
                See example.
            evaluation_window (str): Determines the time window of data to compute the alert
                metric on and is an ISO 8601 formatted duration such
                as "PT1H5M26S".
            evaluation_delay (str): Determines how long the system will wait for data to come
                in before evaluating a window and is an ISO 8601 formatted duration such as
                "PT1H5M26S".
            tags (Optional[Dict[str, str]]): Tags for this alert such as {"env": "dev"}.
            alert_id (Union[str, UUID]): If an update is desired, provide the alert UUID here.

        Returns:
            str: The affected alert ID
        """
        if tags is None:
            tags = {}
        try:
            application_id = self._get_application(application_name)["id"]
        except GantryException as ge:
            raise ValueError(
                f'Could not find "{application_name}". Reason: "{ge}". Check the spelling?'
            )
        _alert_checks = []
        for check in alert_checks:
            check_as_dict = dataclasses.asdict(check)
            check_as_dict["aggregation"] = check_as_dict["aggregation"].value
            check_as_dict["data_node_ids"] = [
                data_node_id
                for __, data_node_id in self._map_column_names_to_ids(
                    application_name, check_as_dict["column_names"]
                ).items()
            ]
            check_as_dict.pop("column_names")
            _alert_checks.append(check_as_dict)
        json_data = {
            "alert_type": GantryAlerts.ALERT_TYPE,
            "alert_name": alert_name,
            "model_node_id": application_id,
            "relative_time": evaluation_window,
            "delay_time": evaluation_delay,
            "checks": _alert_checks,
            "tags": tags,
        }
        if alert_id is not None:
            # This will update instead of create on the backend. Names can be re-used
            # across alert ids so we can't disambiguate one-to-one without it.
            json_data["alert_id"] = alert_id

        response = self._api_client.request(
            "POST", "/api/v1/alerts", json=json_data, raise_for_status=True
        )
        return response["data"]["alert"]["id"]

    def create_alert(
        self,
        application_name: str,
        alert_name: str,
        alert_checks: List[AlertsCheck],
        evaluation_window: str,
        evaluation_delay: str,
        tags: Optional[Dict[str, str]] = None,
    ) -> str:
        """
        Create an alert in your Gantry application.

        NOTE: Alerts will only trigger for data whose tags match the alert tags here,
        if applicable.

        All alerts require check definitions that define what metrics trigger
        the alert. An example check is:

        .. code-block:: python

            from gantry.alerts import AlertsCheck

            check = AlertsCheck(
                aggregation=AlertsAggregation.MAX,
                data_node_names=[
                    "inputs.field",
                ],
                lower_bound=10,
                upper_bound=20,
            )

        Args:
            application_name (str): The name of your existing application in Gantry.
            alert_name (str): Whatever you would like to call your alert!
            alert_checks (List[AlertsCheck]): A list of metric checks that define this alert.
                See example.
            evaluation_window (str): Determines the time window of data to compute the alert
                metric on and is an ISO 8601 formatted duration such
                as "PT1H5M26S".
            evaluation_delay (str): Determines how long the system will wait for data to come
                in before evaluating a window and is an ISO 8601 formatted duration such as
                "PT1H5M26S".
            tags (Optional[Dict[str, str]]): Tags for this alert such as {"env": "dev"}.

        Returns:
            str: The affected alert ID
        """
        return self._create_or_update_alert(
            application_name=application_name,
            alert_name=alert_name,
            alert_checks=alert_checks,
            evaluation_window=evaluation_window,
            evaluation_delay=evaluation_delay,
            tags=tags,
        )

    def update_alert(
        self,
        application_name: str,
        alert_id: Union[str, UUID],
        alert_name: str,
        alert_checks: List[AlertsCheck],
        evaluation_window: str,
        evaluation_delay: str,
        tags: Optional[Dict[str, str]] = None,
    ) -> str:
        """
        Update an alert in your Gantry application.

        NOTE: This method will OVERWRITE your existing alert checks.

        All alerts require check definitions that define what metrics trigger
        the alert. An example check is:

        .. code-block:: python

            from gantry.alerts import AlertsCheck

            check = AlertsCheck(
                aggregation=AlertsAggregation.MAX,
                data_node_names=[
                    "inputs.field",
                ],
                lower_bound=10,
                upper_bound=20,
            )

        Args:
            application_name (str): The name of your existing application in Gantry.
            alert_id (Union[str, UUID]): The ID for your alert. (You can use the .get_alerts()
                method to get the IDs of your existing alerts.)
            alert_name (str): Whatever you would like to call your (existing) alert!
            alert_checks (List[AlertCheck]): A list of metric checks that define this alert.
                See example.
            evaluation_window (str): Determines the time window of data to compute the alert
                metric on and is an ISO 8601 formatted duration such
                as "PT1H5M26S".
            evaluation_delay (str): Determines how long the system will wait for data to come
                in before evaluating a window and is an ISO 8601 formatted duration such as
                "PT1H5M26S".
            tags (Optional[Dict[str, str]]): Tags for this alert such as {"env": "dev"}.

        Returns:
            str: The affected alert ID
        """
        return self._create_or_update_alert(
            application_name=application_name,
            alert_id=alert_id,
            alert_name=alert_name,
            alert_checks=alert_checks,
            evaluation_window=evaluation_window,
            evaluation_delay=evaluation_delay,
            tags=tags,
        )

    # To hide the complexity of the backend data model from the user, use
    # separate create / delete methods for Slack notifications (otherwise
    # they will need the channel_id / frequency_id, and editting these notifications
    # did not work for me in the API anyway.
    def create_slack_notification(
        self,
        alert_id: str,
        notification_name: str,
        slack_webhook_url: str,
        notify_daily: bool,
        daily_notification_time: Optional[str] = None,
    ) -> str:
        """
        An alert notification sets up the means by which you will be notified
        in case any of your alert checks trigger. This method will set up a
        Slack notification for you.

        Args:
            alert_id (Union[str, UUID]): The alert to which this Slack notification will trigger on.
            notification_name (str): The name for this notification.
            slack_webhook_url (str): The webhook for the Slack application.
            notify_daily (bool): Whether or not the notification should occur daily at a particular
                time (see arg `daily_notification_time`).
            daily_notification_time (str): The time at which you will be notified daily if
                `notify_daily` is True, e.g. "12:00:00"

        Returns:
            str: The affected Slack notification ID.
        """
        if notify_daily:
            frequency = SlackNotificationFrequency.DAILY.value
        else:
            frequency = SlackNotificationFrequency.IMMEDIATELY.value
        json_data = {
            "channel_type": GantryAlerts.CHANNEL_TYPE,
            "channel_name": notification_name,
            "webhook_url": slack_webhook_url,
            "alert_id": alert_id,
            "frequency": frequency,
        }
        if daily_notification_time is not None:
            json_data["frequency_relative_time"] = daily_notification_time
        response = self._api_client.request(
            "POST", "/api/v1/alerts/notifications", json=json_data, raise_for_status=True
        )
        return response["data"]["channel"]["id"]

    def get_slack_notifications(self) -> List[Dict]:
        """
        Get a list of all of the Slack notification objects in the system.

        Returns:
            List[Dict]: All of your Slack notification objects.
        """
        response = self._api_client.request("GET", "/api/v1/alerts/channels", raise_for_status=True)
        return response["data"]["channels"]

    def delete_alert(self, alert_id: Union[str, UUID]) -> None:
        """
        Deletes a particular alert, specified by `alert_id`.

        Args:
            alert_id (Union[str, UUID]): The alert_id.

        Returns:
            None: This is a side-effect.
        """
        self._api_client.request("DELETE", f"/api/v1/alerts/{alert_id}", raise_for_status=True)

    def delete_slack_notification(
        self, alert_id: Union[str, UUID], slack_notification_id: Union[str, UUID]
    ) -> None:
        """
        Deletes a particular Slack notification, specified by the alert and
        Slack notification IDs.

        Args:
            alert_id (Union[str, UUID]): The alert ID.
            slack_notification_id (Union[str, UUID]): The Slack notification ID.
        """
        json_data = {"alert_id": alert_id, "channel_id": slack_notification_id}
        self._api_client.request(
            "DELETE", "/api/v1/alerts/channels", json=json_data, raise_for_status=True
        )



import logging
import re
from typing import List

logger = logging.getLogger(__name__)


PROMPT_INPUT_REGEX = re.compile(r"\{\{([^\}\{]+)\}\}")


def _get_prompt_inputs(prompt: str) -> List[str]:
    return list(set(re.findall(PROMPT_INPUT_REGEX, prompt)))  # unique list


def fill_prompt(prompt: str, values: dict, strict: bool = False) -> str:
    """
    Fills a prompt with values, returning the completed prompt.

    Example usage:

    .. code-block:: python

        fill_prompt(
            "This is a prompt template. {{input1}} {{input2}}",
            {"input1": "foo", "input2": "bar"},
        )

    Example output:

    .. code-block:: python

        "This is a prompt template. foo bar"



    Args:
        prompt (str): The prompt template to use for this version. Denote variables
            like this: `{{variable}}`
        values (dict): A key-value mapping of variable names to values. The values will
            be used to fill the prompt template.
        strict (bool): If True, raises a ValueError if the prompt is not fully filled.
            Otherwise, logs a warning. False by default.

    Returns:
        str: The filled prompt.
    """
    unused_keys = []

    for key, value in values.items():
        prompt, num_replaced = re.subn("{{" + key + "}}", value, prompt)
        if num_replaced == 0:
            unused_keys.append(key)

    unfilled_keys = _get_prompt_inputs(prompt)

    if unused_keys or unfilled_keys:
        error_message = (
            f"Prompt {prompt!r} was not fully filled. "
            f"Unused keys: {unused_keys}. "
            f"Unfilled keys: {unfilled_keys}."
        )

        if strict:
            raise ValueError(error_message)
        else:
            logger.warning(error_message)

    return prompt



import logging

import pytest

from gantry.applications.llm_utils import _get_prompt_inputs, fill_prompt


@pytest.mark.parametrize(
    ("prompt", "expected"),
    [
        ("no vars", []),
        ("{{test}} variable", ["test"]),
        ("{{two}} {{vars}}", ["two", "vars"]),
        ("{not} a {variable}", []),
        ("{{test}} {{test}}", ["test"]),
    ],
)
def test_get_prompt_inputs(prompt, expected):
    assert set(_get_prompt_inputs(prompt)) == set(expected)


@pytest.mark.parametrize(
    ("prompt", "values", "expected"),
    [
        ("no vars", {}, "no vars"),
        ("{{test}} variable", {"test": "filled"}, "filled variable"),
        ("{{two}} {{vars}}", {"two": "2", "vars": "variables"}, "2 variables"),
        ("{{unfilled}} var", {}, "{{unfilled}} var"),
        ("{{test}} {{test}}", {"test": "filled"}, "filled filled"),
        ("{{extra}} fields", {"extra": "filled", "unused": "unused"}, "filled fields"),
    ],
)
def test_fill_prompt(prompt, values, expected):
    assert fill_prompt(prompt, values) == expected


@pytest.mark.parametrize(
    ("prompt", "values", "strict", "expected_error"),
    [
        ("{{test}} variable", {}, False, "warn"),
        ("{{test}} variable", {}, True, "error"),
        ("test variable", {"test": "foo"}, False, "warn"),
        ("test variable", {"test": "foo"}, True, "error"),
        ("{{test}} variable", {"test": "foo"}, True, None),
        ("{{test}} variable", {"test": "foo"}, False, None),
    ],
)
def test_fill_prompt_errors_or_warnings(prompt, values, strict, expected_error, caplog):
    if expected_error is None:
        with caplog.at_level(logging.WARNING):
            fill_prompt(prompt, values, strict=strict)
        assert not caplog.text
    elif expected_error == "warn":
        with caplog.at_level(logging.WARNING):
            fill_prompt(prompt, values, strict=strict)
        assert "was not fully filled" in caplog.text
    elif expected_error == "error":
        with pytest.raises(ValueError):
            fill_prompt(prompt, values, strict=strict)


@pytest.mark.parametrize(
    ("prompt", "values", "message"),
    [
        (
            "{{a}} b",
            {},
            "Prompt '{{a}} b' was not fully filled. Unused keys: []. Unfilled keys: ['a'].",
        ),
        (
            "a",
            {"a": "b"},
            "Prompt 'a' was not fully filled. Unused keys: ['a']. Unfilled keys: [].",
        ),
    ],
)
def test_fill_prompt_correct_error_message(prompt, values, message, caplog):
    with caplog.at_level(logging.WARNING):
        fill_prompt(prompt, values)
    assert message in caplog.text



import logging
from abc import ABC
from typing import Optional

from gantry.alerts import (
    create_slack_notification,
    delete_alert,
    delete_slack_notification,
)

logger = logging.getLogger(__name__)


class Action(ABC):
    pass


class SendSlackMessage(Action):
    def __init__(
        self,
        name: str,
        webhook_url: str,
        id: Optional[str] = None,
        notify_daily: bool = False,
        alert_id: Optional[str] = None,
    ):
        self.id = id
        self.name = name
        self.webhook_url = webhook_url
        self.notify_daily = notify_daily
        self.alert_id = alert_id

    def update_trigger(self, alert_id: str):
        self.alert_id = alert_id

    def start(self):
        logger.info("Start sending slack notification upon triggered.")

        self.id = create_slack_notification(
            self.alert_id,
            self.name,
            self.webhook_url,
            notify_daily=self.notify_daily,
        )
        logger.info("Slack ID: " + self.id)

    def to_dict(self):
        return {
            "type": "SendSlackMessage",
            "content": {
                "id": self.id,
                "name": self.name,
                "webhook_url": self.webhook_url,
                "alert_id": self.alert_id,
                "notify_daily": self.notify_daily,
            },
        }

    @classmethod
    def from_dict(cls, d):
        return cls(
            id=d["id"],
            name=d["name"],
            webhook_url=d["webhook_url"],
            notify_daily=d["notify_daily"],
            alert_id=d["alert_id"],
        )

    def stop(self):
        """
        Delete the ongoing slack notification and alert.
        """
        delete_slack_notification(slack_notification_id=self.id, alert_id=self.alert_id)
        delete_alert(self.alert_id)



import functools
import inspect
import logging
import os
from typing import Optional
from urllib.parse import urlparse

from gantry.api_client import APIClient
from gantry.automations.curators import globals
from gantry.automations.curators.curators import CuratorClient
from gantry.const import PROD_API_URL
from gantry.exceptions import ClientNotInitialized

logger_obj = logging.getLogger(__name__)


def _curators_alias(func):
    """Decorator for CuratorClient functions, exposed in main.py"""
    doc = "Alias for :meth:`gantry.automations.curators.curators.CuratorClient.{}`".format(
        func.__name__
    )
    signature = inspect.signature(getattr(CuratorClient, func.__name__))
    orig_doc = inspect.getdoc(getattr(CuratorClient, func.__name__))

    if orig_doc:
        doc += "\n\n{}".format(orig_doc)
    signature = signature.replace(parameters=tuple(signature.parameters.values())[1:])

    func.__doc__ = doc
    func.__signature__ = signature

    # This decorator also checks that the _CURATOR_CLIENT
    # has been initialized
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if globals._CURATOR_CLIENT is None:
            raise ClientNotInitialized()

        return func(*args, **kwargs)

    return wrapper


def _init(
    api_key: Optional[str] = None,
):
    backend = os.environ.get("GANTRY_LOGS_LOCATION", PROD_API_URL)
    parsed_origin = urlparse(backend)
    if parsed_origin.scheme not in ("http", "https"):
        raise ValueError(
            "Invalid backend. http or https backends " + "supported. Got {}".format(backend)
        )

    api_key = os.environ.get("GANTRY_API_KEY") if api_key is None else api_key

    if not api_key:
        raise ValueError(
            """
            No API key provided. Please pass the api_key parameter or set the GANTRY_API_KEY
            environment variable.
            """
        )

    api_client = APIClient(backend, api_key)
    # instantiated curators need a client, too
    globals._API_CLIENT = api_client  # type: ignore
    # this client has special methods for querying all curators
    globals._CURATOR_CLIENT = CuratorClient(api_client)  # type: ignore

    if not globals._CURATOR_CLIENT.ping():  # type: ignore
        logger_obj.warning(
            "Gantry services not reachable. Check provided URL "
            f"[{backend}] is pointing to the correct address"
        )
        return

    if not globals._CURATOR_CLIENT.ready():  # type: ignore
        logger_obj.warning("Gantry services won't receive traffic. Check if API Key is valid")


@_curators_alias
def get_all_curators(*args, **kwargs):
    return globals._CURATOR_CLIENT.get_all_curators(*args, **kwargs)  # type: ignore


@_curators_alias
def get_curator(*args, **kwargs):
    return globals._CURATOR_CLIENT.get_curator(*args, **kwargs)  # type: ignore


@_curators_alias
def list_curators(*args, **kwargs):
    return globals._CURATOR_CLIENT.list_curators(*args, **kwargs)  # type: ignore



import pytest

from gantry.api_client import APIClient
from gantry.applications.client import ApplicationClient

HOST = "https://test-api"


@pytest.fixture
def test_api_client():
    return APIClient(origin=HOST)


@pytest.fixture
def test_client(test_api_client):
    return ApplicationClient(api_client=test_api_client)



class GantryException(Exception):
    """
    Base class for Gantry exceptions.
    """

    pass


class GantryLoggingException(GantryException):
    """
    Raised when logging invalid data.
    """

    pass


class GantryLoggingDataTypeError(GantryLoggingException):
    """
    Raised when Gantry cannot process datatype
    """

    pass


class ClientNotInitialized(GantryException):
    """
    Gantry client not initialized
    """

    def __init__(self):
        message = "Gantry client is not initialized. Did you call `gantry.init()`?"
        super(ClientNotInitialized, self).__init__(message)


class QueryError(GantryException):
    """
    Raised when a query doesn't have a result
    """

    pass


class GantryBatchCreationException(GantryException):
    """
    Raised when gantry failed to create batch
    """

    pass


class GantryRequestException(GantryException):
    """
    Raised when an API request returns an error.
    """

    def __init__(self, url: str, status_code: int, response_text: str) -> None:
        if status_code == 404:
            if "/api/v1/ingest/data-connectors" in url:
                msg = (
                    "Data connector exception. Ensure the parameters refer to an existing "
                    "data connector you have access to. Details: {}"
                ).format(response_text)
            else:
                msg = (
                    "A resource was not found while trying to access ({}). Ensure the parameters "
                    "refer to an existing resource you have access to. Details: {}"
                ).format(url, response_text)
        elif status_code == 429:
            if "/api/v1/ingest/data-connectors" in url:
                msg = (
                    "Data connector has hit the concurrency limit. Please reach out to Gantry "
                    + "support if you would like to raise this limit."
                )
            else:
                msg = (
                    "Logger has hit the rate limit. Please reach out to Gantry "
                    "support if you would like to raise this limit."
                )
        elif status_code == 401:
            msg = (
                "Authentication error. Ensure that you supplied a working API key by "
                "calling gantry.ready()."
            )
        elif status_code == 403:
            msg = (
                "Access denied. Check with your organization admin or Gantry support "
                "that you have the appropriate permissions to perform this action"
            )
        elif status_code >= 400 and status_code < 500:
            msg = (
                "Malformed data error. Ensure Gantry is receiving valid data. Details: {}"
            ).format(response_text)
        else:
            msg = "Internal Gantry error. Contact Gantry support for help."

        err_msg = "{} (Status code: {})".format(msg, status_code)
        self.status_code = status_code
        super().__init__(err_msg)


class DatasetDeletedException(GantryException):
    """
    Raised when an illegal dataset operation is attempted on a deleted dataset
    """

    pass


class CuratorNotFoundError(GantryException):
    """
    Raised when a curator is not found and thus cannot be enabled / disabled
    """

    pass


class DatasetCommitNotFoundError(GantryException):
    """
    Raised when a specific requested dataset commit is not found
    """

    pass


class DatasetNotFoundError(GantryException):
    """
    Raised when a specific requested dataset is not found
    """

    pass


class DatasetHeadOutOfDateException(GantryException):
    """
    Raised when a the head is out of date for a dataset operation
    """

    def __init__(self):
        message = "Local HEAD not up to date! Your local version is behind the remote"
        super().__init__(message)


class NoTabularDataError(GantryException):
    """
    Raised when there is no tabular data folder in a dataset
    """

    def __init__(self):
        message = (
            "No tabular data found in tabular_manifests folder! Please make sure that the"
            "tabular_manifests folder exists and contains at least one manifest file."
        )
        super().__init__(message)


class DataSchemaMismatchError(GantryException):
    """
    Raised when the dataset config schema disagrees with the csv content
    """

    def __init__(self):
        message = (
            "The dataset config schema disagrees with the data! "
            "Ensure that the columns in the dataset config match those in the tabular data files "
            "and that the types in the config are all correct."
        )
        super().__init__(message)
